/* token.c - Tokenizer - Don Yang / Dave Newquist / John Rapp

   07/26/01
                                                                           */

#include<ctype.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include"token.h"


/* Local functions */
static int GetByte(Input *input);
static Token *GetEndTag(Input *input, Token *token);
static Token *GetTag(Input *input, Token *token);
static Token *GetText(Input *input, Token *token);
static int PeekByte(Input *input);

/************************************************************* DeleteInput */
void DeleteInput(Input *input)
{
   if( input )
      free(input);

} /* DeleteInput() */

/************************************************************* DeleteToken */
void DeleteToken(Token *token)
{
   if( token )
   {
      if( token->type == TokenText )
         free(token->lexeme.s);
      free(token);
   }

} /* DeleteToken() */

/**************************************************************** GetToken */
Token *GetToken(Input *input)
{
   Token *ret;
   int c;

   if( (ret = (Token*)malloc(sizeof(Token))) == NULL )
   {
      fputs("Out of memory\n", stderr);
      return NULL;
   }
   ret->lexeme.s = NULL;
   ret->tagtype = 0;

   c = PeekByte(input);
   switch( c )
   {
      case '>':
         fputs("Syntax error\n", stderr);
         free(ret);
         return NULL;

      case '<':
         GetByte(input);
         return GetTag(input, ret);

      case EOF:
         GetByte(input);
         ret->type = TokenEOF;
         break;

      case '\t':
      case '\r':
      case '\n':
      case ' ':
         ret->type = TokenSpace;
         ret->lexeme.i = GetByte(input);
         break;

      default:
         ret->type = TokenText;
         return GetText(input, ret);
   }

   return ret;

} /* GetToken() */

/*************************************************************** InitInput */
Input *InitInput(FILE *file)
{
   Input *ret;

   if( (ret = (Input*)malloc(sizeof(Input))) == NULL )
   {
      fputs("Not enough memory\n", stderr);
      return NULL;
   }

   ret->file = file;
   ret->index = ret->bufsize = 0;
   return ret;

} /* InitInput() */


/***************************************************************** GetByte */
static int GetByte(Input *input)
{
   size_t size;

   if( input->index >= input->bufsize )
   {
      if( !(size = fread(input->buffer, 1, INPUT_BUFFER_SIZE, input->file)) )
         return EOF;
      input->bufsize = (int)size;
      input->index = 0;
   }

   return (int)input->buffer[input->index++];

} /* GetByte() */

/*************************************************************** GetEndTag */
static Token *GetEndTag(Input *input, Token *token)
{
   if( GetByte(input) != '>' )
   {
      fputs("Syntax error in tag\n", stderr);
      free(token);
      return NULL;
   }
   return token;

} /* GetEndTag() */

/****************************************************************** GetTag */
static Token *GetTag(Input *input, Token *token)
{
   int c;

   if( PeekByte(input) == '/' )
   {
      GetByte(input);
      token->tagtype = TagClose;
   }
   else
   {
      token->tagtype = TagOpen;
   }

   /* First character */
   c = GetByte(input);
   if( isdigit(c) )
   {
      token->type = TagSize;
      token->lexeme.i = c;
      return GetEndTag(input, token);
   }
   switch( c )
   {
      case 'B': token->type = TagB; return GetEndTag(input, token);
      case 'I': token->type = TagI; return GetEndTag(input, token);
      case 'S': token->type = TagS; return GetEndTag(input, token);
      case 'U': token->type = TagU; return GetEndTag(input, token);

      case 'r': case 'g': case 'b': case 'w':
      case 'c': case 'm': case 'y': case 'k':
         token->type = TagColor;
         token->lexeme.i = c;
         return GetEndTag(input, token);

      /* Second character */
      case 'E':
         if( GetByte(input) == 'M' )
         {
            token->type = TagEM;
            return GetEndTag(input, token);
         }
         break;

      case 'P':
         if( GetByte(input) == 'L' )
         {
            token->type = TagPL;
            return GetEndTag(input, token);
         }
         break;

      case 'T':
         if( GetByte(input) == 'T' )
         {
            token->type = TagTT;
            return GetEndTag(input, token);
         }
         break;

      default:
         break;
   }

   fputs("Syntax error in tag\n", stderr);
   free(token);
   return NULL;

} /* GetTag() */

/***************************************************************** GetText */
static Token *GetText(Input *input, Token *token)
{
   char *lexeme, *tmp;
   size_t i, size;
   int c;

   size = 32;
   if( (lexeme = (char*)malloc(size)) == NULL )
   {
      fputs("Out of memory\n", stderr);
      free(token);
      return NULL;
   }

   i = 0;
   while( (c = PeekByte(input)) != EOF )
   {
      if( c == '<' || c == '\r' || c == '\n' || c == '\t' || c == ' ' )
         break;
      if( c == '>' )
      {
         fputs("Syntax error\n", stderr);
         free(lexeme);
         free(token);
         return NULL;
      }
      c = GetByte(input);
      lexeme[i++] = (char)c;
      if( i + 2 >= size )
      {
         if( (tmp = (char*)malloc(size *= 2)) == NULL )
         {
            fputs("Out of memory\n", stderr);
            free(lexeme);
            free(token);
            return NULL;
         }
         memcpy(tmp, lexeme, i);
         free(lexeme);
         lexeme = tmp;
      }
   }

   lexeme[i] = 0;
   token->type = TokenText;
   token->lexeme.s = lexeme;
   return token;

} /* GetText() */

/**************************************************************** PeekByte */
static int PeekByte(Input *input)
{
   size_t size;

   if( input->index >= input->bufsize )
   {
      if( !(size = fread(input->buffer, 1, INPUT_BUFFER_SIZE, input->file)) )
         return EOF;
      input->bufsize = (int)size;
      input->index = 0;
   }

   return (int)input->buffer[input->index];

} /* PeekByte() */
